% --------------------------------------------------------------------------------------------------
% Apply OMP-F, OMP-G and SOMP-F algorithms to the frames of an audio signal  
% Random Holes case 
% Implemented by: Toumi.I
% October 2017
% Some parameters to set:
% ---------- h: Hop size 
% ---------- N: size of frame
% ---------- NbFram: Number of similar frames to be used for the sparse Non Local sparse algorithm 
% ---------------------------------------------------------------------------------------------------

clear all
close all

setpath;

%% Read test signal

[FileName, FilePath] = uigetfile(('*.wav'),'Select the audio file');
myAudio=fullfile(FilePath, FileName);
[sig,Fs] = audioread(myAudio);

%% Initialization

N = 256;                 % Size of one frame
ls = size(sig,1);        % Size of the signal    
h = N/4;                 % hop size (step)
n0List = 1:h:ls-N+1;     % List of the Frames start's indexes 
L0 = length(n0List);     % Number of the frames

%% Dictionary Generation

D = iDFT_Dict(N,2);
DG = Gabor_Dict(N,2);

%% Initialization of the algorithm

E2M = 1e-6;              % Tolerance for the algos
T = round(N/4);          % maximum number of coefficients

%% Generate the problem

% Select the percentage of missing data 

percentage = inputdlg({'Missing data (%):'}, '', 1, {'0'});
valper = str2double(percentage{1,1});

if (valper>100)||(valper<0)||(mod(valper,1)~=0)
    error('Percentage must be an integer compromised between 0 and 100 ');
end

percentage=valper/100;

holeSize=2;                                 % Size of the holes 
NMissSamples=round(ls*percentage);          % Number of the missing samples
NHoles = NMissSamples/holeSize;             % Number of the holes

problemParameters.NHoles = round(NHoles);
problemParameters.holeSize = holeSize;

[problemData, solutionData] = generateMissingGroupsProblem(sig,problemParameters);
fprintf('%d - %d-length holes: %.2f missing samples\n',NHoles,holeSize,NMissSamples/ls*100);

sign=solutionData.xClean;                   % Normalized original signal

Mask = zeros(N,L0);     % Mask with (0 / 1) for (observed / missing) samples
S = zeros(N,L0);        % Matrix of all the frames form the original signal 
Sm = zeros(N,L0);       % Matrix of all the frames with missing samples 

% Mask Computation

for l=1:L0
    n = n0List(l);
    S(:,l) = sign(n+(0:N-1));          
    
    sn = problemData.x(n+(0:N-1));
    Sm(:,l) = sn;                      
    
    Mask(:,l)= problemData.IMiss(n+(0:N-1));
end


%% Initialization

S_OMP = zeros(N,L0);            % Reconstructed frames from OMP
S_OMPG = zeros(N,L0);           % Reconstructed frames from OMPG
S_SOMP = zeros(N,L0);           % Reconstructed frames from S-OMP
    
MSEALL_OMP = zeros(1,L0);       % MSE computed for the frames on all samples - OMP
MSEALL_OMPG = zeros(1,L0);      % MSE computed for the frames on all samples - OMPG
MSEALL_SOMP = zeros(1,L0);      % MSE computed for the frames on all samples - SOMP

MSEMISS_OMP = zeros(1,L0);      % MSE computed for the frames on missing samples - OMP
MSEMISS_OMPG = zeros(1,L0);     % MSE computed for the frames on missing samples - OMPG
MSEMISS_SOMP = zeros(1,L0);     % MSE computed for the frames on missing samples - SOMP

SNRALL_OMP = zeros(1,L0);       % SNR computed for the frames on all samples - OMP
SNRALL_OMPG = zeros(1,L0);      % SNR computed for the frames on all samples - OMPG
SNRALL_SOMP = zeros(1,L0);      % SNR computed for the frames on all samples - SOMP

SNRMISS_OMP = zeros(1,L0);      % SNR computed for the frames on missing samples - OMP
SNRMISS_OMPG = zeros(1,L0);     % SNR computed for the frames on missing samples - OMPG
SNRMISS_SOMP = zeros(1,L0);     % SNR computed for the frames on missing samples - SOMP

NbFram = 4;                     % Number of similar frames to be selected for joint sparsity

%% Apply OMP, OMPG and S-OMP to all the frames

for l = 1:L0     % for each frame
    
    
    % ---- Frame selection -----
    
    sn = Sm(:,l) ;              % Select the frame 
    IObs = find(~Mask(:,l));    % Select the indexes of observed samples in the frame
    IMiss = find(Mask(:,l));    % Select the indexes of missing samples in the frame
    
    % ---- Similar frames selection  ( IS distance - Cos|.| distance) -----
    
%     d_IS = FFTSimSignal(sn,problemData.x,h,'ISDist');
%     
%     [AIS,BIS] = sort(d_IS);             % Sort the distance in ascending order
%     IndxISList = BIS(1:NbFram);
%     S_IS = Sm(:,IndxISList);   

    gamma_abs = FFTSimSignal(sn,problemData.x,h,'CosDistAbs');

    [Aabs,Babs] = sort(1-gamma_abs);      % Sort the distance in ascending order
    IndxabsList = Babs(1:NbFram);         % Select indexes of only the NbFram similar frames
    
    S_abs = Sm(:,IndxabsList);            % Matrix of the set of similar frames
    
    % ---- Normalize dictionary for OMP & OMP G ----
    % Taking into account only the observed samples in the considered frame 
    
    Dict = D(IObs,:);
    W = 1./sqrt(diag(Dict'*Dict));    
    Dict = Dict*diag(W);
    
    DictG = DG(IObs,:);
    WG = 1./sqrt(diag(DictG'*DictG));    
    DictG = DictG*diag(WG);
    
    % ------ OMP algorithm ------
    
    xObs = sn(IObs);                            % selection of observed samples in the frame 
    [a_omp, indxomp] = OMP(xObs,Dict,T,E2M);    % Inpainting OMP algorithm 
    S_OMP(:,l) = D*(W.*a_omp);                  % Reconstruct the frame with OMP      
    
    % -------- OMPG algorithm ---------
    
    [a_ompg, indxG] = OMP_Gabor(xObs,DictG,T,E2M);  % Inpainting OMP-Gabor algorithm
    S_OMPG(:,l) = DG*(WG.*a_ompg);                  % Reconstruct the frame with OMPG
    
    % -------- SOMP algorithm ---------

%     X = S_IS; 
%     MaskX = Mask(:,IndxISList);
    
    X = S_abs;                                  % Matrix of the selected similar frames 
    MaskX = Mask(:,IndxabsList);                % Matrix of the masks corresponding to the set of similar frames 
    
    [A_somp,indxsomp] = SOMP_Inpainting(X,D,MaskX,T,E2M);       % S-OMP algorithm
    S_SOMP(:,l) = D*(W.*A_somp(:,1));                           % Reconstruct the frame with S-OMP
    % P.S: we take the first column of A since the first frame is the similar set is the selected frame                                       
    % The normalization vector W is computed according to the observed samples in this frame 
    
    % --------- Evaluation ---------- 
    
    MSEALL_OMP(l) = MSE(S(:,l),S_OMP(:,l));
    MSEALL_OMPG(l) = MSE(S(:,l),S_OMPG(:,l));
    MSEALL_SOMP(l) = MSE(S(:,l),S_SOMP(:,l));

    SNRALL_OMP(l) = SNR(S(:,l),S_OMP(:,l));
    SNRALL_OMPG(l) = SNR(S(:,l),S_OMPG(:,l));
    SNRALL_SOMP(l) = SNR(S(:,l),S_SOMP(:,l));

    if ~ isempty(IMiss)
        MSEMISS_OMP(l) = MSE(S(IMiss,l),S_OMP(IMiss));
        MSEMISS_OMPG(l) = MSE(S(IMiss,l),S_OMPG(IMiss,l));
        MSEMISS_SOMP(l) = MSE(S(IMiss,l),S_SOMP(IMiss,l));
    
        SNRMISS_OMP(l) = SNR(S(IMiss,l),S_OMP(IMiss,l));
        SNRMISS_OMPG(l) = SNR(S(IMiss,l),S_OMPG(IMiss,l));
        SNRMISS_SOMP(l) = SNR(S(IMiss,l),S_SOMP(IMiss,l));
    end
    fprintf('End frame processing n: %d\n',l);
end

% ------ Average MSE / SNR through all the frames--------------

MSE_OMP_allsamples= mean(MSEALL_OMP);
MSE_OMP_missingsamples= mean(MSEMISS_OMP);

MSE_OMPG_allsamples= mean(MSEALL_OMPG);
MSE_OMPG_missingsamples= mean(MSEMISS_OMPG);

MSE_SOMP_allsamples= mean(MSEALL_SOMP);
MSE_SOMP_missingsamples= mean(MSEMISS_SOMP);

SNR_OMP_allsamples= mean(SNRALL_OMP);
SNR_OMP_missingsamples= mean(SNRMISS_OMP);

SNR_OMPG_allsamples= mean(SNRALL_OMPG);
SNR_OMPG_missingsamples= mean(SNRMISS_OMPG);

SNR_SOMP_allsamples= mean(SNRALL_SOMP);
SNR_SOMP_missingsamples= mean(SNRMISS_SOMP);

% Save results

file_out =['Inpainting_' FileName(1:length(FileName)-4) '_RandomHoles_' num2str(valper) 'percent','.mat'];

save(file_out,'h','N','holeSize','NMissSamples','NbFram','S_OMP','S_OMPG','S_SOMP',...
    'MSE_OMP_allsamples','MSE_OMP_missingsamples',...
    'MSE_OMPG_allsamples','MSE_OMPG_missingsamples','MSE_SOMP_allsamples',...
    'MSE_SOMP_missingsamples','SNR_OMPG_allsamples','SNR_OMPG_missingsamples',...
    'SNR_SOMP_allsamples','SNR_SOMP_missingsamples','SNR_OMP_allsamples','SNR_OMP_missingsamples');
    